@@ -1,17 +1,20 @@
-from enum import Enum
-from ast import parse
-
-from .type_inference import *
-from .rewrite_tuple_assign import RewriteTupleAssign
-from .rewrite_augassign import RewriteAugAssign
-from .rewrite_import_plutusdata import RewriteImportPlutusData
-from .rewrite_import_dataclasses import RewriteImportDataclasses
-from .rewrite_import_typing import RewriteImportTyping
-from .rewrite_import import RewriteImport
-
 import pluthon as plt
 import uplc
 
+from .rewrite.rewrite_augassign import RewriteAugAssign
+from .rewrite.rewrite_import import RewriteImport
+from .rewrite.rewrite_import_dataclasses import RewriteImportDataclasses
+from .rewrite.rewrite_import_plutusdata import RewriteImportPlutusData
+from .rewrite.rewrite_import_typing import RewriteImportTyping
+from .rewrite.rewrite_inject_builtins import RewriteInjectBuiltins
+from .rewrite.rewrite_remove_type_stuff import RewriteRemoveTypeStuff
+from .rewrite.rewrite_tuple_assign import RewriteTupleAssign
+from .optimize.optimize_remove_pass import OptimizeRemovePass
+from .optimize.optimize_remove_deadvars import OptimizeRemoveDeadvars
+from .optimize.optimize_varlen import OptimizeVarlen
+from .type_inference import *
+from .util import RawPlutoExpr, TypedNodeTransformer
+
 STATEMONAD = "s"
 
 
@@ -177,41 +180,17 @@ def extend_statemonad(
     values: typing.List[plt.AST],
     old_statemonad: plt.FunctionalMap,
 ):
-    return plt.FunctionalMapExtend(old_statemonad, [n.encode() for n in names], values)
+    return plt.FunctionalMapExtend(old_statemonad, [n for n in names], values)
 
 
-class PythonBuiltIn(Enum):
-    print = plt.Lambda(
-        ["x", STATEMONAD],
-        plt.Trace(plt.Var("x"), plt.NoneData()),
-    )
-    range = plt.Lambda(
-        ["limit", STATEMONAD],
-        plt.Range(plt.Var("limit")),
-    )
+INITIAL_STATE = plt.FunctionalMap()
 
 
-INITIAL_STATE = plt.FunctionalMapExtend(
-    plt.FunctionalMap(),
-    [b.name for b in PythonBuiltIn],
-    [b.value for b in PythonBuiltIn],
-)
-
-
-class UPLCCompiler(NodeTransformer):
+class UPLCCompiler(TypedNodeTransformer):
     """
     Expects a TypedAST and returns UPLC/Pluto like code
     """
 
-    def visit(self, node):
-        """Visit a node."""
-        node_class_name = node.__class__.__name__
-        if node_class_name.startswith("Typed"):
-            node_class_name = node_class_name[len("Typed") :]
-        method = "visit_" + node_class_name
-        visitor = getattr(self, method, self.generic_visit)
-        return visitor(node)
-
     def visit_sequence(self, node_seq: typing.List[typedstmt]) -> plt.AST:
         s = plt.Var(STATEMONAD)
         for n in node_seq:
@@ -270,7 +249,7 @@ class UPLCCompiler(NodeTransformer):
         # TODO can use more sophisiticated procedure here i.e. functions marked by comment
         main_fun: typing.Optional[InstanceType] = None
         for s in node.body:
-            if isinstance(s, FunctionDef) and s.name == "validator":
+            if isinstance(s, FunctionDef) and s.orig_name == "validator":
                 main_fun = s
         assert main_fun is not None, "Could not find function named validator"
         main_fun_typ: FunctionType = main_fun.typ.typ
@@ -294,7 +273,7 @@ class UPLCCompiler(NodeTransformer):
                                 "g",
                                 plt.FunctionalMapAccess(
                                     plt.Var("s"),
-                                    plt.ByteString("validator".encode("utf8")),
+                                    plt.ByteString(main_fun.name),
                                     plt.TraceError("NameError: validator"),
                                 ),
                             ),
@@ -352,8 +331,8 @@ class UPLCCompiler(NodeTransformer):
                 [STATEMONAD],
                 plt.FunctionalMapAccess(
                     plt.Var(STATEMONAD),
-                    plt.ByteString(node.id.encode()),
-                    plt.TraceError(f"NameError: {node.id}"),
+                    plt.ByteString(node.id),
+                    plt.TraceError(f"NameError: {node.orig_id}"),
                 ),
             )
         raise NotImplementedError(f"Context {node.ctx} not supported")
@@ -586,6 +565,9 @@ class UPLCCompiler(NodeTransformer):
             ),
         )
 
+    def visit_RawPlutoExpr(self, node: RawPlutoExpr) -> plt.AST:
+        return node.expr
+
     def generic_visit(self, node: AST) -> plt.AST:
         raise NotImplementedError(f"Can not compile {node}")
 
@@ -594,13 +576,22 @@ def compile(prog: AST):
     compiler_steps = [
         # Important to call this one first - it imports all further files
         RewriteImport,
-        # The remaining order is almost arbitrary
+        # Rewrites that simplify the python code
         RewriteAugAssign,
         RewriteTupleAssign,
         RewriteImportPlutusData,
         RewriteImportTyping,
         RewriteImportDataclasses,
+        # The type inference needs to be run after complex python operations were rewritten
         AggressiveTypeInferencer,
+        # Rewrites that circumvent the type inference or use its results
+        RewriteRemoveTypeStuff,
+        RewriteInjectBuiltins,
+        # Apply optimizations
+        OptimizeRemoveDeadvars,
+        OptimizeVarlen,
+        OptimizeRemovePass,
+        # the compiler runs last
         UPLCCompiler,
     ]
     for s in compiler_steps:
