@@ -1,9 +1,10 @@
 from copy import copy
-from ast import *
+
+from .typed_ast import *
+from .util import PythonBuiltInTypes
 
 # from frozendict import frozendict
 
-from .typed_ast import *
 
 """
 An aggressive type inference based on the work of Aycock [1].
@@ -32,14 +33,14 @@ INITIAL_SCOPE = dict(
         "Union": UnionType(NoneInstanceType),
         "dataclass": NoneInstanceType,
         "PlutusData": NoneInstanceType,
+    }
+)
+
+INITIAL_SCOPE.update(
+    {
         # builtin functions
-        "print": InstanceType(FunctionType([StringInstanceType], NoneInstanceType)),
-        "range": InstanceType(
-            FunctionType(
-                [IntegerInstanceType],
-                InstanceType(ListType(IntegerInstanceType)),
-            )
-        ),
+        k.name: v
+        for k, v in PythonBuiltInTypes.items()
     }
 )
 
@@ -308,6 +309,7 @@ class AggressiveTypeInferencer(NodeTransformer):
 
     def visit_FunctionDef(self, node: FunctionDef) -> TypedFunctionDef:
         tfd = copy(node)
+        assert not node.decorator_list, "Functions may not have decorators"
         self.enter_scope()
         tfd.args = self.visit(node.args)
         functyp = FunctionType(
